/* DO NOT EDIT - This file generated automatically by gl_gentable.py (from Mesa) script */

/*
 * Copyright (C) 1999-2001  Brian Paul   All Rights Reserved.
 * (C) Copyright IBM Corporation 2004, 2005
 * (C) Copyright Apple Inc 2011
 * All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  IN NO EVENT SHALL
 * BRIAN PAUL, IBM,
 * AND/OR THEIR SUPPLIERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF
 * OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

/* GLXEXT is the define used in the xserver when the GLX extension is being
 * built.  Hijack this to determine whether this file is being built for the
 * server or the client.
 */
#ifdef HAVE_DIX_CONFIG_H
#include <dix-config.h>
#endif

#ifndef _WIN32
#include <dlfcn.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "glapi.h"
#include "glapitable.h"

#ifdef GLXEXT
#include "os.h"
#endif

static void
__glapi_gentable_NoOp(void) {
#if defined(GLXEXT)
    LogMessage(X_ERROR, "GLX: Call to unimplemented API: Unknown\n");
#else
    fprintf(stderr, "Call to unimplemented API: Unknown\n");
#endif
}

static void
__glapi_gentable_set_remaining_noop(struct _glapi_table *disp) {
    GLuint entries = _glapi_get_dispatch_table_size();
    void **dispatch = (void **) disp;
    unsigned i;

    /* ISO C is annoying sometimes */
    union {_glapi_proc p; void *v;} p;
    p.p = __glapi_gentable_NoOp;

    for(i=0; i < entries; i++)
        if(dispatch[i] == NULL)
            dispatch[i] = p.v;
}


#define GLAPI_TABLE_COUNT 1671
static const char * const _glapi_table_func_names[GLAPI_TABLE_COUNT] = {
    /*     0 */ "NewList",
    /*     1 */ "EndList",
    /*     2 */ "CallList",
    /*     3 */ "CallLists",
    /*     4 */ "DeleteLists",
    /*     5 */ "GenLists",
    /*     6 */ "ListBase",
    /*     7 */ "Begin",
    /*     8 */ "Bitmap",
    /*     9 */ "Color3b",
    /*    10 */ "Color3bv",
    /*    11 */ "Color3d",
    /*    12 */ "Color3dv",
    /*    13 */ "Color3f",
    /*    14 */ "Color3fv",
    /*    15 */ "Color3i",
    /*    16 */ "Color3iv",
    /*    17 */ "Color3s",
    /*    18 */ "Color3sv",
    /*    19 */ "Color3ub",
    /*    20 */ "Color3ubv",
    /*    21 */ "Color3ui",
    /*    22 */ "Color3uiv",
    /*    23 */ "Color3us",
    /*    24 */ "Color3usv",
    /*    25 */ "Color4b",
    /*    26 */ "Color4bv",
    /*    27 */ "Color4d",
    /*    28 */ "Color4dv",
    /*    29 */ "Color4f",
    /*    30 */ "Color4fv",
    /*    31 */ "Color4i",
    /*    32 */ "Color4iv",
    /*    33 */ "Color4s",
    /*    34 */ "Color4sv",
    /*    35 */ "Color4ub",
    /*    36 */ "Color4ubv",
    /*    37 */ "Color4ui",
    /*    38 */ "Color4uiv",
    /*    39 */ "Color4us",
    /*    40 */ "Color4usv",
    /*    41 */ "EdgeFlag",
    /*    42 */ "EdgeFlagv",
    /*    43 */ "End",
    /*    44 */ "Indexd",
    /*    45 */ "Indexdv",
    /*    46 */ "Indexf",
    /*    47 */ "Indexfv",
    /*    48 */ "Indexi",
    /*    49 */ "Indexiv",
    /*    50 */ "Indexs",
    /*    51 */ "Indexsv",
    /*    52 */ "Normal3b",
    /*    53 */ "Normal3bv",
    /*    54 */ "Normal3d",
    /*    55 */ "Normal3dv",
    /*    56 */ "Normal3f",
    /*    57 */ "Normal3fv",
    /*    58 */ "Normal3i",
    /*    59 */ "Normal3iv",
    /*    60 */ "Normal3s",
    /*    61 */ "Normal3sv",
    /*    62 */ "RasterPos2d",
    /*    63 */ "RasterPos2dv",
    /*    64 */ "RasterPos2f",
    /*    65 */ "RasterPos2fv",
    /*    66 */ "RasterPos2i",
    /*    67 */ "RasterPos2iv",
    /*    68 */ "RasterPos2s",
    /*    69 */ "RasterPos2sv",
    /*    70 */ "RasterPos3d",
    /*    71 */ "RasterPos3dv",
    /*    72 */ "RasterPos3f",
    /*    73 */ "RasterPos3fv",
    /*    74 */ "RasterPos3i",
    /*    75 */ "RasterPos3iv",
    /*    76 */ "RasterPos3s",
    /*    77 */ "RasterPos3sv",
    /*    78 */ "RasterPos4d",
    /*    79 */ "RasterPos4dv",
    /*    80 */ "RasterPos4f",
    /*    81 */ "RasterPos4fv",
    /*    82 */ "RasterPos4i",
    /*    83 */ "RasterPos4iv",
    /*    84 */ "RasterPos4s",
    /*    85 */ "RasterPos4sv",
    /*    86 */ "Rectd",
    /*    87 */ "Rectdv",
    /*    88 */ "Rectf",
    /*    89 */ "Rectfv",
    /*    90 */ "Recti",
    /*    91 */ "Rectiv",
    /*    92 */ "Rects",
    /*    93 */ "Rectsv",
    /*    94 */ "TexCoord1d",
    /*    95 */ "TexCoord1dv",
    /*    96 */ "TexCoord1f",
    /*    97 */ "TexCoord1fv",
    /*    98 */ "TexCoord1i",
    /*    99 */ "TexCoord1iv",
    /*   100 */ "TexCoord1s",
    /*   101 */ "TexCoord1sv",
    /*   102 */ "TexCoord2d",
    /*   103 */ "TexCoord2dv",
    /*   104 */ "TexCoord2f",
    /*   105 */ "TexCoord2fv",
    /*   106 */ "TexCoord2i",
    /*   107 */ "TexCoord2iv",
    /*   108 */ "TexCoord2s",
    /*   109 */ "TexCoord2sv",
    /*   110 */ "TexCoord3d",
    /*   111 */ "TexCoord3dv",
    /*   112 */ "TexCoord3f",
    /*   113 */ "TexCoord3fv",
    /*   114 */ "TexCoord3i",
    /*   115 */ "TexCoord3iv",
    /*   116 */ "TexCoord3s",
    /*   117 */ "TexCoord3sv",
    /*   118 */ "TexCoord4d",
    /*   119 */ "TexCoord4dv",
    /*   120 */ "TexCoord4f",
    /*   121 */ "TexCoord4fv",
    /*   122 */ "TexCoord4i",
    /*   123 */ "TexCoord4iv",
    /*   124 */ "TexCoord4s",
    /*   125 */ "TexCoord4sv",
    /*   126 */ "Vertex2d",
    /*   127 */ "Vertex2dv",
    /*   128 */ "Vertex2f",
    /*   129 */ "Vertex2fv",
    /*   130 */ "Vertex2i",
    /*   131 */ "Vertex2iv",
    /*   132 */ "Vertex2s",
    /*   133 */ "Vertex2sv",
    /*   134 */ "Vertex3d",
    /*   135 */ "Vertex3dv",
    /*   136 */ "Vertex3f",
    /*   137 */ "Vertex3fv",
    /*   138 */ "Vertex3i",
    /*   139 */ "Vertex3iv",
    /*   140 */ "Vertex3s",
    /*   141 */ "Vertex3sv",
    /*   142 */ "Vertex4d",
    /*   143 */ "Vertex4dv",
    /*   144 */ "Vertex4f",
    /*   145 */ "Vertex4fv",
    /*   146 */ "Vertex4i",
    /*   147 */ "Vertex4iv",
    /*   148 */ "Vertex4s",
    /*   149 */ "Vertex4sv",
    /*   150 */ "ClipPlane",
    /*   151 */ "ColorMaterial",
    /*   152 */ "CullFace",
    /*   153 */ "Fogf",
    /*   154 */ "Fogfv",
    /*   155 */ "Fogi",
    /*   156 */ "Fogiv",
    /*   157 */ "FrontFace",
    /*   158 */ "Hint",
    /*   159 */ "Lightf",
    /*   160 */ "Lightfv",
    /*   161 */ "Lighti",
    /*   162 */ "Lightiv",
    /*   163 */ "LightModelf",
    /*   164 */ "LightModelfv",
    /*   165 */ "LightModeli",
    /*   166 */ "LightModeliv",
    /*   167 */ "LineStipple",
    /*   168 */ "LineWidth",
    /*   169 */ "Materialf",
    /*   170 */ "Materialfv",
    /*   171 */ "Materiali",
    /*   172 */ "Materialiv",
    /*   173 */ "PointSize",
    /*   174 */ "PolygonMode",
    /*   175 */ "PolygonStipple",
    /*   176 */ "Scissor",
    /*   177 */ "ShadeModel",
    /*   178 */ "TexParameterf",
    /*   179 */ "TexParameterfv",
    /*   180 */ "TexParameteri",
    /*   181 */ "TexParameteriv",
    /*   182 */ "TexImage1D",
    /*   183 */ "TexImage2D",
    /*   184 */ "TexEnvf",
    /*   185 */ "TexEnvfv",
    /*   186 */ "TexEnvi",
    /*   187 */ "TexEnviv",
    /*   188 */ "TexGend",
    /*   189 */ "TexGendv",
    /*   190 */ "TexGenf",
    /*   191 */ "TexGenfv",
    /*   192 */ "TexGeni",
    /*   193 */ "TexGeniv",
    /*   194 */ "FeedbackBuffer",
    /*   195 */ "SelectBuffer",
    /*   196 */ "RenderMode",
    /*   197 */ "InitNames",
    /*   198 */ "LoadName",
    /*   199 */ "PassThrough",
    /*   200 */ "PopName",
    /*   201 */ "PushName",
    /*   202 */ "DrawBuffer",
    /*   203 */ "Clear",
    /*   204 */ "ClearAccum",
    /*   205 */ "ClearIndex",
    /*   206 */ "ClearColor",
    /*   207 */ "ClearStencil",
    /*   208 */ "ClearDepth",
    /*   209 */ "StencilMask",
    /*   210 */ "ColorMask",
    /*   211 */ "DepthMask",
    /*   212 */ "IndexMask",
    /*   213 */ "Accum",
    /*   214 */ "Disable",
    /*   215 */ "Enable",
    /*   216 */ "Finish",
    /*   217 */ "Flush",
    /*   218 */ "PopAttrib",
    /*   219 */ "PushAttrib",
    /*   220 */ "Map1d",
    /*   221 */ "Map1f",
    /*   222 */ "Map2d",
    /*   223 */ "Map2f",
    /*   224 */ "MapGrid1d",
    /*   225 */ "MapGrid1f",
    /*   226 */ "MapGrid2d",
    /*   227 */ "MapGrid2f",
    /*   228 */ "EvalCoord1d",
    /*   229 */ "EvalCoord1dv",
    /*   230 */ "EvalCoord1f",
    /*   231 */ "EvalCoord1fv",
    /*   232 */ "EvalCoord2d",
    /*   233 */ "EvalCoord2dv",
    /*   234 */ "EvalCoord2f",
    /*   235 */ "EvalCoord2fv",
    /*   236 */ "EvalMesh1",
    /*   237 */ "EvalPoint1",
    /*   238 */ "EvalMesh2",
    /*   239 */ "EvalPoint2",
    /*   240 */ "AlphaFunc",
    /*   241 */ "BlendFunc",
    /*   242 */ "LogicOp",
    /*   243 */ "StencilFunc",
    /*   244 */ "StencilOp",
    /*   245 */ "DepthFunc",
    /*   246 */ "PixelZoom",
    /*   247 */ "PixelTransferf",
    /*   248 */ "PixelTransferi",
    /*   249 */ "PixelStoref",
    /*   250 */ "PixelStorei",
    /*   251 */ "PixelMapfv",
    /*   252 */ "PixelMapuiv",
    /*   253 */ "PixelMapusv",
    /*   254 */ "ReadBuffer",
    /*   255 */ "CopyPixels",
    /*   256 */ "ReadPixels",
    /*   257 */ "DrawPixels",
    /*   258 */ "GetBooleanv",
    /*   259 */ "GetClipPlane",
    /*   260 */ "GetDoublev",
    /*   261 */ "GetError",
    /*   262 */ "GetFloatv",
    /*   263 */ "GetIntegerv",
    /*   264 */ "GetLightfv",
    /*   265 */ "GetLightiv",
    /*   266 */ "GetMapdv",
    /*   267 */ "GetMapfv",
    /*   268 */ "GetMapiv",
    /*   269 */ "GetMaterialfv",
    /*   270 */ "GetMaterialiv",
    /*   271 */ "GetPixelMapfv",
    /*   272 */ "GetPixelMapuiv",
    /*   273 */ "GetPixelMapusv",
    /*   274 */ "GetPolygonStipple",
    /*   275 */ "GetString",
    /*   276 */ "GetTexEnvfv",
    /*   277 */ "GetTexEnviv",
    /*   278 */ "GetTexGendv",
    /*   279 */ "GetTexGenfv",
    /*   280 */ "GetTexGeniv",
    /*   281 */ "GetTexImage",
    /*   282 */ "GetTexParameterfv",
    /*   283 */ "GetTexParameteriv",
    /*   284 */ "GetTexLevelParameterfv",
    /*   285 */ "GetTexLevelParameteriv",
    /*   286 */ "IsEnabled",
    /*   287 */ "IsList",
    /*   288 */ "DepthRange",
    /*   289 */ "Frustum",
    /*   290 */ "LoadIdentity",
    /*   291 */ "LoadMatrixf",
    /*   292 */ "LoadMatrixd",
    /*   293 */ "MatrixMode",
    /*   294 */ "MultMatrixf",
    /*   295 */ "MultMatrixd",
    /*   296 */ "Ortho",
    /*   297 */ "PopMatrix",
    /*   298 */ "PushMatrix",
    /*   299 */ "Rotated",
    /*   300 */ "Rotatef",
    /*   301 */ "Scaled",
    /*   302 */ "Scalef",
    /*   303 */ "Translated",
    /*   304 */ "Translatef",
    /*   305 */ "Viewport",
    /*   306 */ "ArrayElement",
    /*   307 */ "BindTexture",
    /*   308 */ "ColorPointer",
    /*   309 */ "DisableClientState",
    /*   310 */ "DrawArrays",
    /*   311 */ "DrawElements",
    /*   312 */ "EdgeFlagPointer",
    /*   313 */ "EnableClientState",
    /*   314 */ "IndexPointer",
    /*   315 */ "Indexub",
    /*   316 */ "Indexubv",
    /*   317 */ "InterleavedArrays",
    /*   318 */ "NormalPointer",
    /*   319 */ "PolygonOffset",
    /*   320 */ "TexCoordPointer",
    /*   321 */ "VertexPointer",
    /*   322 */ "AreTexturesResident",
    /*   323 */ "CopyTexImage1D",
    /*   324 */ "CopyTexImage2D",
    /*   325 */ "CopyTexSubImage1D",
    /*   326 */ "CopyTexSubImage2D",
    /*   327 */ "DeleteTextures",
    /*   328 */ "GenTextures",
    /*   329 */ "GetPointerv",
    /*   330 */ "IsTexture",
    /*   331 */ "PrioritizeTextures",
    /*   332 */ "TexSubImage1D",
    /*   333 */ "TexSubImage2D",
    /*   334 */ "PopClientAttrib",
    /*   335 */ "PushClientAttrib",
    /*   336 */ "BlendColor",
    /*   337 */ "BlendEquation",
    /*   338 */ "DrawRangeElements",
    /*   339 */ "ColorTable",
    /*   340 */ "ColorTableParameterfv",
    /*   341 */ "ColorTableParameteriv",
    /*   342 */ "CopyColorTable",
    /*   343 */ "GetColorTable",
    /*   344 */ "GetColorTableParameterfv",
    /*   345 */ "GetColorTableParameteriv",
    /*   346 */ "ColorSubTable",
    /*   347 */ "CopyColorSubTable",
    /*   348 */ "ConvolutionFilter1D",
    /*   349 */ "ConvolutionFilter2D",
    /*   350 */ "ConvolutionParameterf",
    /*   351 */ "ConvolutionParameterfv",
    /*   352 */ "ConvolutionParameteri",
    /*   353 */ "ConvolutionParameteriv",
    /*   354 */ "CopyConvolutionFilter1D",
    /*   355 */ "CopyConvolutionFilter2D",
    /*   356 */ "GetConvolutionFilter",
    /*   357 */ "GetConvolutionParameterfv",
    /*   358 */ "GetConvolutionParameteriv",
    /*   359 */ "GetSeparableFilter",
    /*   360 */ "SeparableFilter2D",
    /*   361 */ "GetHistogram",
    /*   362 */ "GetHistogramParameterfv",
    /*   363 */ "GetHistogramParameteriv",
    /*   364 */ "GetMinmax",
    /*   365 */ "GetMinmaxParameterfv",
    /*   366 */ "GetMinmaxParameteriv",
    /*   367 */ "Histogram",
    /*   368 */ "Minmax",
    /*   369 */ "ResetHistogram",
    /*   370 */ "ResetMinmax",
    /*   371 */ "TexImage3D",
    /*   372 */ "TexSubImage3D",
    /*   373 */ "CopyTexSubImage3D",
    /*   374 */ "ActiveTexture",
    /*   375 */ "ClientActiveTexture",
    /*   376 */ "MultiTexCoord1d",
    /*   377 */ "MultiTexCoord1dv",
    /*   378 */ "MultiTexCoord1fARB",
    /*   379 */ "MultiTexCoord1fvARB",
    /*   380 */ "MultiTexCoord1i",
    /*   381 */ "MultiTexCoord1iv",
    /*   382 */ "MultiTexCoord1s",
    /*   383 */ "MultiTexCoord1sv",
    /*   384 */ "MultiTexCoord2d",
    /*   385 */ "MultiTexCoord2dv",
    /*   386 */ "MultiTexCoord2fARB",
    /*   387 */ "MultiTexCoord2fvARB",
    /*   388 */ "MultiTexCoord2i",
    /*   389 */ "MultiTexCoord2iv",
    /*   390 */ "MultiTexCoord2s",
    /*   391 */ "MultiTexCoord2sv",
    /*   392 */ "MultiTexCoord3d",
    /*   393 */ "MultiTexCoord3dv",
    /*   394 */ "MultiTexCoord3fARB",
    /*   395 */ "MultiTexCoord3fvARB",
    /*   396 */ "MultiTexCoord3i",
    /*   397 */ "MultiTexCoord3iv",
    /*   398 */ "MultiTexCoord3s",
    /*   399 */ "MultiTexCoord3sv",
    /*   400 */ "MultiTexCoord4d",
    /*   401 */ "MultiTexCoord4dv",
    /*   402 */ "MultiTexCoord4fARB",
    /*   403 */ "MultiTexCoord4fvARB",
    /*   404 */ "MultiTexCoord4i",
    /*   405 */ "MultiTexCoord4iv",
    /*   406 */ "MultiTexCoord4s",
    /*   407 */ "MultiTexCoord4sv",
    /*   408 */ "CompressedTexImage1D",
    /*   409 */ "CompressedTexImage2D",
    /*   410 */ "CompressedTexImage3D",
    /*   411 */ "CompressedTexSubImage1D",
    /*   412 */ "CompressedTexSubImage2D",
    /*   413 */ "CompressedTexSubImage3D",
    /*   414 */ "GetCompressedTexImage",
    /*   415 */ "LoadTransposeMatrixd",
    /*   416 */ "LoadTransposeMatrixf",
    /*   417 */ "MultTransposeMatrixd",
    /*   418 */ "MultTransposeMatrixf",
    /*   419 */ "SampleCoverage",
    /*   420 */ "BlendFuncSeparate",
    /*   421 */ "FogCoordPointer",
    /*   422 */ "FogCoordd",
    /*   423 */ "FogCoorddv",
    /*   424 */ "MultiDrawArrays",
    /*   425 */ "PointParameterf",
    /*   426 */ "PointParameterfv",
    /*   427 */ "PointParameteri",
    /*   428 */ "PointParameteriv",
    /*   429 */ "SecondaryColor3b",
    /*   430 */ "SecondaryColor3bv",
    /*   431 */ "SecondaryColor3d",
    /*   432 */ "SecondaryColor3dv",
    /*   433 */ "SecondaryColor3i",
    /*   434 */ "SecondaryColor3iv",
    /*   435 */ "SecondaryColor3s",
    /*   436 */ "SecondaryColor3sv",
    /*   437 */ "SecondaryColor3ub",
    /*   438 */ "SecondaryColor3ubv",
    /*   439 */ "SecondaryColor3ui",
    /*   440 */ "SecondaryColor3uiv",
    /*   441 */ "SecondaryColor3us",
    /*   442 */ "SecondaryColor3usv",
    /*   443 */ "SecondaryColorPointer",
    /*   444 */ "WindowPos2d",
    /*   445 */ "WindowPos2dv",
    /*   446 */ "WindowPos2f",
    /*   447 */ "WindowPos2fv",
    /*   448 */ "WindowPos2i",
    /*   449 */ "WindowPos2iv",
    /*   450 */ "WindowPos2s",
    /*   451 */ "WindowPos2sv",
    /*   452 */ "WindowPos3d",
    /*   453 */ "WindowPos3dv",
    /*   454 */ "WindowPos3f",
    /*   455 */ "WindowPos3fv",
    /*   456 */ "WindowPos3i",
    /*   457 */ "WindowPos3iv",
    /*   458 */ "WindowPos3s",
    /*   459 */ "WindowPos3sv",
    /*   460 */ "BeginQuery",
    /*   461 */ "BindBuffer",
    /*   462 */ "BufferData",
    /*   463 */ "BufferSubData",
    /*   464 */ "DeleteBuffers",
    /*   465 */ "DeleteQueries",
    /*   466 */ "EndQuery",
    /*   467 */ "GenBuffers",
    /*   468 */ "GenQueries",
    /*   469 */ "GetBufferParameteriv",
    /*   470 */ "GetBufferPointerv",
    /*   471 */ "GetBufferSubData",
    /*   472 */ "GetQueryObjectiv",
    /*   473 */ "GetQueryObjectuiv",
    /*   474 */ "GetQueryiv",
    /*   475 */ "IsBuffer",
    /*   476 */ "IsQuery",
    /*   477 */ "MapBuffer",
    /*   478 */ "UnmapBuffer",
    /*   479 */ "AttachShader",
    /*   480 */ "BindAttribLocation",
    /*   481 */ "BlendEquationSeparate",
    /*   482 */ "CompileShader",
    /*   483 */ "CreateProgram",
    /*   484 */ "CreateShader",
    /*   485 */ "DeleteProgram",
    /*   486 */ "DeleteShader",
    /*   487 */ "DetachShader",
    /*   488 */ "DisableVertexAttribArray",
    /*   489 */ "DrawBuffers",
    /*   490 */ "EnableVertexAttribArray",
    /*   491 */ "GetActiveAttrib",
    /*   492 */ "GetActiveUniform",
    /*   493 */ "GetAttachedShaders",
    /*   494 */ "GetAttribLocation",
    /*   495 */ "GetProgramInfoLog",
    /*   496 */ "GetProgramiv",
    /*   497 */ "GetShaderInfoLog",
    /*   498 */ "GetShaderSource",
    /*   499 */ "GetShaderiv",
    /*   500 */ "GetUniformLocation",
    /*   501 */ "GetUniformfv",
    /*   502 */ "GetUniformiv",
    /*   503 */ "GetVertexAttribPointerv",
    /*   504 */ "GetVertexAttribdv",
    /*   505 */ "GetVertexAttribfv",
    /*   506 */ "GetVertexAttribiv",
    /*   507 */ "IsProgram",
    /*   508 */ "IsShader",
    /*   509 */ "LinkProgram",
    /*   510 */ "ShaderSource",
    /*   511 */ "StencilFuncSeparate",
    /*   512 */ "StencilMaskSeparate",
    /*   513 */ "StencilOpSeparate",
    /*   514 */ "Uniform1f",
    /*   515 */ "Uniform1fv",
    /*   516 */ "Uniform1i",
    /*   517 */ "Uniform1iv",
    /*   518 */ "Uniform2f",
    /*   519 */ "Uniform2fv",
    /*   520 */ "Uniform2i",
    /*   521 */ "Uniform2iv",
    /*   522 */ "Uniform3f",
    /*   523 */ "Uniform3fv",
    /*   524 */ "Uniform3i",
    /*   525 */ "Uniform3iv",
    /*   526 */ "Uniform4f",
    /*   527 */ "Uniform4fv",
    /*   528 */ "Uniform4i",
    /*   529 */ "Uniform4iv",
    /*   530 */ "UniformMatrix2fv",
    /*   531 */ "UniformMatrix3fv",
    /*   532 */ "UniformMatrix4fv",
    /*   533 */ "UseProgram",
    /*   534 */ "ValidateProgram",
    /*   535 */ "VertexAttrib1d",
    /*   536 */ "VertexAttrib1dv",
    /*   537 */ "VertexAttrib1s",
    /*   538 */ "VertexAttrib1sv",
    /*   539 */ "VertexAttrib2d",
    /*   540 */ "VertexAttrib2dv",
    /*   541 */ "VertexAttrib2s",
    /*   542 */ "VertexAttrib2sv",
    /*   543 */ "VertexAttrib3d",
    /*   544 */ "VertexAttrib3dv",
    /*   545 */ "VertexAttrib3s",
    /*   546 */ "VertexAttrib3sv",
    /*   547 */ "VertexAttrib4Nbv",
    /*   548 */ "VertexAttrib4Niv",
    /*   549 */ "VertexAttrib4Nsv",
    /*   550 */ "VertexAttrib4Nub",
    /*   551 */ "VertexAttrib4Nubv",
    /*   552 */ "VertexAttrib4Nuiv",
    /*   553 */ "VertexAttrib4Nusv",
    /*   554 */ "VertexAttrib4bv",
    /*   555 */ "VertexAttrib4d",
    /*   556 */ "VertexAttrib4dv",
    /*   557 */ "VertexAttrib4iv",
    /*   558 */ "VertexAttrib4s",
    /*   559 */ "VertexAttrib4sv",
    /*   560 */ "VertexAttrib4ubv",
    /*   561 */ "VertexAttrib4uiv",
    /*   562 */ "VertexAttrib4usv",
    /*   563 */ "VertexAttribPointer",
    /*   564 */ "UniformMatrix2x3fv",
    /*   565 */ "UniformMatrix2x4fv",
    /*   566 */ "UniformMatrix3x2fv",
    /*   567 */ "UniformMatrix3x4fv",
    /*   568 */ "UniformMatrix4x2fv",
    /*   569 */ "UniformMatrix4x3fv",
    /*   570 */ "BeginConditionalRender",
    /*   571 */ "BeginTransformFeedback",
    /*   572 */ "BindBufferBase",
    /*   573 */ "BindBufferRange",
    /*   574 */ "BindFragDataLocation",
    /*   575 */ "ClampColor",
    /*   576 */ "ClearBufferfi",
    /*   577 */ "ClearBufferfv",
    /*   578 */ "ClearBufferiv",
    /*   579 */ "ClearBufferuiv",
    /*   580 */ "ColorMaski",
    /*   581 */ "Disablei",
    /*   582 */ "Enablei",
    /*   583 */ "EndConditionalRender",
    /*   584 */ "EndTransformFeedback",
    /*   585 */ "GetBooleani_v",
    /*   586 */ "GetFragDataLocation",
    /*   587 */ "GetIntegeri_v",
    /*   588 */ "GetStringi",
    /*   589 */ "GetTexParameterIiv",
    /*   590 */ "GetTexParameterIuiv",
    /*   591 */ "GetTransformFeedbackVarying",
    /*   592 */ "GetUniformuiv",
    /*   593 */ "GetVertexAttribIiv",
    /*   594 */ "GetVertexAttribIuiv",
    /*   595 */ "IsEnabledi",
    /*   596 */ "TexParameterIiv",
    /*   597 */ "TexParameterIuiv",
    /*   598 */ "TransformFeedbackVaryings",
    /*   599 */ "Uniform1ui",
    /*   600 */ "Uniform1uiv",
    /*   601 */ "Uniform2ui",
    /*   602 */ "Uniform2uiv",
    /*   603 */ "Uniform3ui",
    /*   604 */ "Uniform3uiv",
    /*   605 */ "Uniform4ui",
    /*   606 */ "Uniform4uiv",
    /*   607 */ "VertexAttribI1iv",
    /*   608 */ "VertexAttribI1uiv",
    /*   609 */ "VertexAttribI4bv",
    /*   610 */ "VertexAttribI4sv",
    /*   611 */ "VertexAttribI4ubv",
    /*   612 */ "VertexAttribI4usv",
    /*   613 */ "VertexAttribIPointer",
    /*   614 */ "PrimitiveRestartIndex",
    /*   615 */ "TexBuffer",
    /*   616 */ "FramebufferTexture",
    /*   617 */ "GetBufferParameteri64v",
    /*   618 */ "GetInteger64i_v",
    /*   619 */ "VertexAttribDivisor",
    /*   620 */ "MinSampleShading",
    /*   621 */ "MemoryBarrierByRegion",
    /*   622 */ "BindProgramARB",
    /*   623 */ "DeleteProgramsARB",
    /*   624 */ "GenProgramsARB",
    /*   625 */ "GetProgramEnvParameterdvARB",
    /*   626 */ "GetProgramEnvParameterfvARB",
    /*   627 */ "GetProgramLocalParameterdvARB",
    /*   628 */ "GetProgramLocalParameterfvARB",
    /*   629 */ "GetProgramStringARB",
    /*   630 */ "GetProgramivARB",
    /*   631 */ "IsProgramARB",
    /*   632 */ "ProgramEnvParameter4dARB",
    /*   633 */ "ProgramEnvParameter4dvARB",
    /*   634 */ "ProgramEnvParameter4fARB",
    /*   635 */ "ProgramEnvParameter4fvARB",
    /*   636 */ "ProgramLocalParameter4dARB",
    /*   637 */ "ProgramLocalParameter4dvARB",
    /*   638 */ "ProgramLocalParameter4fARB",
    /*   639 */ "ProgramLocalParameter4fvARB",
    /*   640 */ "ProgramStringARB",
    /*   641 */ "VertexAttrib1fARB",
    /*   642 */ "VertexAttrib1fvARB",
    /*   643 */ "VertexAttrib2fARB",
    /*   644 */ "VertexAttrib2fvARB",
    /*   645 */ "VertexAttrib3fARB",
    /*   646 */ "VertexAttrib3fvARB",
    /*   647 */ "VertexAttrib4fARB",
    /*   648 */ "VertexAttrib4fvARB",
    /*   649 */ "AttachObjectARB",
    /*   650 */ "CreateProgramObjectARB",
    /*   651 */ "CreateShaderObjectARB",
    /*   652 */ "DeleteObjectARB",
    /*   653 */ "DetachObjectARB",
    /*   654 */ "GetAttachedObjectsARB",
    /*   655 */ "GetHandleARB",
    /*   656 */ "GetInfoLogARB",
    /*   657 */ "GetObjectParameterfvARB",
    /*   658 */ "GetObjectParameterivARB",
    /*   659 */ "DrawArraysInstanced",
    /*   660 */ "DrawElementsInstanced",
    /*   661 */ "BindFramebuffer",
    /*   662 */ "BindRenderbuffer",
    /*   663 */ "BlitFramebuffer",
    /*   664 */ "CheckFramebufferStatus",
    /*   665 */ "DeleteFramebuffers",
    /*   666 */ "DeleteRenderbuffers",
    /*   667 */ "FramebufferRenderbuffer",
    /*   668 */ "FramebufferTexture1D",
    /*   669 */ "FramebufferTexture2D",
    /*   670 */ "FramebufferTexture3D",
    /*   671 */ "FramebufferTextureLayer",
    /*   672 */ "GenFramebuffers",
    /*   673 */ "GenRenderbuffers",
    /*   674 */ "GenerateMipmap",
    /*   675 */ "GetFramebufferAttachmentParameteriv",
    /*   676 */ "GetRenderbufferParameteriv",
    /*   677 */ "IsFramebuffer",
    /*   678 */ "IsRenderbuffer",
    /*   679 */ "RenderbufferStorage",
    /*   680 */ "RenderbufferStorageMultisample",
    /*   681 */ "FlushMappedBufferRange",
    /*   682 */ "MapBufferRange",
    /*   683 */ "BindVertexArray",
    /*   684 */ "DeleteVertexArrays",
    /*   685 */ "GenVertexArrays",
    /*   686 */ "IsVertexArray",
    /*   687 */ "GetActiveUniformBlockName",
    /*   688 */ "GetActiveUniformBlockiv",
    /*   689 */ "GetActiveUniformName",
    /*   690 */ "GetActiveUniformsiv",
    /*   691 */ "GetUniformBlockIndex",
    /*   692 */ "GetUniformIndices",
    /*   693 */ "UniformBlockBinding",
    /*   694 */ "CopyBufferSubData",
    /*   695 */ "ClientWaitSync",
    /*   696 */ "DeleteSync",
    /*   697 */ "FenceSync",
    /*   698 */ "GetInteger64v",
    /*   699 */ "GetSynciv",
    /*   700 */ "IsSync",
    /*   701 */ "WaitSync",
    /*   702 */ "DrawElementsBaseVertex",
    /*   703 */ "DrawElementsInstancedBaseVertex",
    /*   704 */ "DrawRangeElementsBaseVertex",
    /*   705 */ "MultiDrawElementsBaseVertex",
    /*   706 */ "ProvokingVertex",
    /*   707 */ "GetMultisamplefv",
    /*   708 */ "SampleMaski",
    /*   709 */ "TexImage2DMultisample",
    /*   710 */ "TexImage3DMultisample",
    /*   711 */ "BlendEquationSeparateiARB",
    /*   712 */ "BlendEquationiARB",
    /*   713 */ "BlendFuncSeparateiARB",
    /*   714 */ "BlendFunciARB",
    /*   715 */ "BindFragDataLocationIndexed",
    /*   716 */ "GetFragDataIndex",
    /*   717 */ "BindSampler",
    /*   718 */ "DeleteSamplers",
    /*   719 */ "GenSamplers",
    /*   720 */ "GetSamplerParameterIiv",
    /*   721 */ "GetSamplerParameterIuiv",
    /*   722 */ "GetSamplerParameterfv",
    /*   723 */ "GetSamplerParameteriv",
    /*   724 */ "IsSampler",
    /*   725 */ "SamplerParameterIiv",
    /*   726 */ "SamplerParameterIuiv",
    /*   727 */ "SamplerParameterf",
    /*   728 */ "SamplerParameterfv",
    /*   729 */ "SamplerParameteri",
    /*   730 */ "SamplerParameteriv",
    /*   731 */ "GetQueryObjecti64v",
    /*   732 */ "GetQueryObjectui64v",
    /*   733 */ "QueryCounter",
    /*   734 */ "ColorP3ui",
    /*   735 */ "ColorP3uiv",
    /*   736 */ "ColorP4ui",
    /*   737 */ "ColorP4uiv",
    /*   738 */ "MultiTexCoordP1ui",
    /*   739 */ "MultiTexCoordP1uiv",
    /*   740 */ "MultiTexCoordP2ui",
    /*   741 */ "MultiTexCoordP2uiv",
    /*   742 */ "MultiTexCoordP3ui",
    /*   743 */ "MultiTexCoordP3uiv",
    /*   744 */ "MultiTexCoordP4ui",
    /*   745 */ "MultiTexCoordP4uiv",
    /*   746 */ "NormalP3ui",
    /*   747 */ "NormalP3uiv",
    /*   748 */ "SecondaryColorP3ui",
    /*   749 */ "SecondaryColorP3uiv",
    /*   750 */ "TexCoordP1ui",
    /*   751 */ "TexCoordP1uiv",
    /*   752 */ "TexCoordP2ui",
    /*   753 */ "TexCoordP2uiv",
    /*   754 */ "TexCoordP3ui",
    /*   755 */ "TexCoordP3uiv",
    /*   756 */ "TexCoordP4ui",
    /*   757 */ "TexCoordP4uiv",
    /*   758 */ "VertexAttribP1ui",
    /*   759 */ "VertexAttribP1uiv",
    /*   760 */ "VertexAttribP2ui",
    /*   761 */ "VertexAttribP2uiv",
    /*   762 */ "VertexAttribP3ui",
    /*   763 */ "VertexAttribP3uiv",
    /*   764 */ "VertexAttribP4ui",
    /*   765 */ "VertexAttribP4uiv",
    /*   766 */ "VertexP2ui",
    /*   767 */ "VertexP2uiv",
    /*   768 */ "VertexP3ui",
    /*   769 */ "VertexP3uiv",
    /*   770 */ "VertexP4ui",
    /*   771 */ "VertexP4uiv",
    /*   772 */ "DrawArraysIndirect",
    /*   773 */ "DrawElementsIndirect",
    /*   774 */ "GetUniformdv",
    /*   775 */ "Uniform1d",
    /*   776 */ "Uniform1dv",
    /*   777 */ "Uniform2d",
    /*   778 */ "Uniform2dv",
    /*   779 */ "Uniform3d",
    /*   780 */ "Uniform3dv",
    /*   781 */ "Uniform4d",
    /*   782 */ "Uniform4dv",
    /*   783 */ "UniformMatrix2dv",
    /*   784 */ "UniformMatrix2x3dv",
    /*   785 */ "UniformMatrix2x4dv",
    /*   786 */ "UniformMatrix3dv",
    /*   787 */ "UniformMatrix3x2dv",
    /*   788 */ "UniformMatrix3x4dv",
    /*   789 */ "UniformMatrix4dv",
    /*   790 */ "UniformMatrix4x2dv",
    /*   791 */ "UniformMatrix4x3dv",
    /*   792 */ "GetActiveSubroutineName",
    /*   793 */ "GetActiveSubroutineUniformName",
    /*   794 */ "GetActiveSubroutineUniformiv",
    /*   795 */ "GetProgramStageiv",
    /*   796 */ "GetSubroutineIndex",
    /*   797 */ "GetSubroutineUniformLocation",
    /*   798 */ "GetUniformSubroutineuiv",
    /*   799 */ "UniformSubroutinesuiv",
    /*   800 */ "PatchParameterfv",
    /*   801 */ "PatchParameteri",
    /*   802 */ "BindTransformFeedback",
    /*   803 */ "DeleteTransformFeedbacks",
    /*   804 */ "DrawTransformFeedback",
    /*   805 */ "GenTransformFeedbacks",
    /*   806 */ "IsTransformFeedback",
    /*   807 */ "PauseTransformFeedback",
    /*   808 */ "ResumeTransformFeedback",
    /*   809 */ "BeginQueryIndexed",
    /*   810 */ "DrawTransformFeedbackStream",
    /*   811 */ "EndQueryIndexed",
    /*   812 */ "GetQueryIndexediv",
    /*   813 */ "ClearDepthf",
    /*   814 */ "DepthRangef",
    /*   815 */ "GetShaderPrecisionFormat",
    /*   816 */ "ReleaseShaderCompiler",
    /*   817 */ "ShaderBinary",
    /*   818 */ "GetProgramBinary",
    /*   819 */ "ProgramBinary",
    /*   820 */ "ProgramParameteri",
    /*   821 */ "GetVertexAttribLdv",
    /*   822 */ "VertexAttribL1d",
    /*   823 */ "VertexAttribL1dv",
    /*   824 */ "VertexAttribL2d",
    /*   825 */ "VertexAttribL2dv",
    /*   826 */ "VertexAttribL3d",
    /*   827 */ "VertexAttribL3dv",
    /*   828 */ "VertexAttribL4d",
    /*   829 */ "VertexAttribL4dv",
    /*   830 */ "VertexAttribLPointer",
    /*   831 */ "DepthRangeArrayv",
    /*   832 */ "DepthRangeIndexed",
    /*   833 */ "GetDoublei_v",
    /*   834 */ "GetFloati_v",
    /*   835 */ "ScissorArrayv",
    /*   836 */ "ScissorIndexed",
    /*   837 */ "ScissorIndexedv",
    /*   838 */ "ViewportArrayv",
    /*   839 */ "ViewportIndexedf",
    /*   840 */ "ViewportIndexedfv",
    /*   841 */ "GetGraphicsResetStatusARB",
    /*   842 */ "GetnColorTableARB",
    /*   843 */ "GetnCompressedTexImageARB",
    /*   844 */ "GetnConvolutionFilterARB",
    /*   845 */ "GetnHistogramARB",
    /*   846 */ "GetnMapdvARB",
    /*   847 */ "GetnMapfvARB",
    /*   848 */ "GetnMapivARB",
    /*   849 */ "GetnMinmaxARB",
    /*   850 */ "GetnPixelMapfvARB",
    /*   851 */ "GetnPixelMapuivARB",
    /*   852 */ "GetnPixelMapusvARB",
    /*   853 */ "GetnPolygonStippleARB",
    /*   854 */ "GetnSeparableFilterARB",
    /*   855 */ "GetnTexImageARB",
    /*   856 */ "GetnUniformdvARB",
    /*   857 */ "GetnUniformfvARB",
    /*   858 */ "GetnUniformivARB",
    /*   859 */ "GetnUniformuivARB",
    /*   860 */ "ReadnPixelsARB",
    /*   861 */ "DrawArraysInstancedBaseInstance",
    /*   862 */ "DrawElementsInstancedBaseInstance",
    /*   863 */ "DrawElementsInstancedBaseVertexBaseInstance",
    /*   864 */ "DrawTransformFeedbackInstanced",
    /*   865 */ "DrawTransformFeedbackStreamInstanced",
    /*   866 */ "GetInternalformativ",
    /*   867 */ "GetActiveAtomicCounterBufferiv",
    /*   868 */ "BindImageTexture",
    /*   869 */ "MemoryBarrier",
    /*   870 */ "TexStorage1D",
    /*   871 */ "TexStorage2D",
    /*   872 */ "TexStorage3D",
    /*   873 */ "TextureStorage1DEXT",
    /*   874 */ "TextureStorage2DEXT",
    /*   875 */ "TextureStorage3DEXT",
    /*   876 */ "ClearBufferData",
    /*   877 */ "ClearBufferSubData",
    /*   878 */ "DispatchCompute",
    /*   879 */ "DispatchComputeIndirect",
    /*   880 */ "CopyImageSubData",
    /*   881 */ "TextureView",
    /*   882 */ "BindVertexBuffer",
    /*   883 */ "VertexAttribBinding",
    /*   884 */ "VertexAttribFormat",
    /*   885 */ "VertexAttribIFormat",
    /*   886 */ "VertexAttribLFormat",
    /*   887 */ "VertexBindingDivisor",
    /*   888 */ "FramebufferParameteri",
    /*   889 */ "GetFramebufferParameteriv",
    /*   890 */ "GetInternalformati64v",
    /*   891 */ "MultiDrawArraysIndirect",
    /*   892 */ "MultiDrawElementsIndirect",
    /*   893 */ "GetProgramInterfaceiv",
    /*   894 */ "GetProgramResourceIndex",
    /*   895 */ "GetProgramResourceLocation",
    /*   896 */ "GetProgramResourceLocationIndex",
    /*   897 */ "GetProgramResourceName",
    /*   898 */ "GetProgramResourceiv",
    /*   899 */ "ShaderStorageBlockBinding",
    /*   900 */ "TexBufferRange",
    /*   901 */ "TexStorage2DMultisample",
    /*   902 */ "TexStorage3DMultisample",
    /*   903 */ "BufferStorage",
    /*   904 */ "ClearTexImage",
    /*   905 */ "ClearTexSubImage",
    /*   906 */ "BindBuffersBase",
    /*   907 */ "BindBuffersRange",
    /*   908 */ "BindImageTextures",
    /*   909 */ "BindSamplers",
    /*   910 */ "BindTextures",
    /*   911 */ "BindVertexBuffers",
    /*   912 */ "GetImageHandleARB",
    /*   913 */ "GetTextureHandleARB",
    /*   914 */ "GetTextureSamplerHandleARB",
    /*   915 */ "GetVertexAttribLui64vARB",
    /*   916 */ "IsImageHandleResidentARB",
    /*   917 */ "IsTextureHandleResidentARB",
    /*   918 */ "MakeImageHandleNonResidentARB",
    /*   919 */ "MakeImageHandleResidentARB",
    /*   920 */ "MakeTextureHandleNonResidentARB",
    /*   921 */ "MakeTextureHandleResidentARB",
    /*   922 */ "ProgramUniformHandleui64ARB",
    /*   923 */ "ProgramUniformHandleui64vARB",
    /*   924 */ "UniformHandleui64ARB",
    /*   925 */ "UniformHandleui64vARB",
    /*   926 */ "VertexAttribL1ui64ARB",
    /*   927 */ "VertexAttribL1ui64vARB",
    /*   928 */ "DispatchComputeGroupSizeARB",
    /*   929 */ "MultiDrawArraysIndirectCountARB",
    /*   930 */ "MultiDrawElementsIndirectCountARB",
    /*   931 */ "ClipControl",
    /*   932 */ "BindTextureUnit",
    /*   933 */ "BlitNamedFramebuffer",
    /*   934 */ "CheckNamedFramebufferStatus",
    /*   935 */ "ClearNamedBufferData",
    /*   936 */ "ClearNamedBufferSubData",
    /*   937 */ "ClearNamedFramebufferfi",
    /*   938 */ "ClearNamedFramebufferfv",
    /*   939 */ "ClearNamedFramebufferiv",
    /*   940 */ "ClearNamedFramebufferuiv",
    /*   941 */ "CompressedTextureSubImage1D",
    /*   942 */ "CompressedTextureSubImage2D",
    /*   943 */ "CompressedTextureSubImage3D",
    /*   944 */ "CopyNamedBufferSubData",
    /*   945 */ "CopyTextureSubImage1D",
    /*   946 */ "CopyTextureSubImage2D",
    /*   947 */ "CopyTextureSubImage3D",
    /*   948 */ "CreateBuffers",
    /*   949 */ "CreateFramebuffers",
    /*   950 */ "CreateProgramPipelines",
    /*   951 */ "CreateQueries",
    /*   952 */ "CreateRenderbuffers",
    /*   953 */ "CreateSamplers",
    /*   954 */ "CreateTextures",
    /*   955 */ "CreateTransformFeedbacks",
    /*   956 */ "CreateVertexArrays",
    /*   957 */ "DisableVertexArrayAttrib",
    /*   958 */ "EnableVertexArrayAttrib",
    /*   959 */ "FlushMappedNamedBufferRange",
    /*   960 */ "GenerateTextureMipmap",
    /*   961 */ "GetCompressedTextureImage",
    /*   962 */ "GetNamedBufferParameteri64v",
    /*   963 */ "GetNamedBufferParameteriv",
    /*   964 */ "GetNamedBufferPointerv",
    /*   965 */ "GetNamedBufferSubData",
    /*   966 */ "GetNamedFramebufferAttachmentParameteriv",
    /*   967 */ "GetNamedFramebufferParameteriv",
    /*   968 */ "GetNamedRenderbufferParameteriv",
    /*   969 */ "GetQueryBufferObjecti64v",
    /*   970 */ "GetQueryBufferObjectiv",
    /*   971 */ "GetQueryBufferObjectui64v",
    /*   972 */ "GetQueryBufferObjectuiv",
    /*   973 */ "GetTextureImage",
    /*   974 */ "GetTextureLevelParameterfv",
    /*   975 */ "GetTextureLevelParameteriv",
    /*   976 */ "GetTextureParameterIiv",
    /*   977 */ "GetTextureParameterIuiv",
    /*   978 */ "GetTextureParameterfv",
    /*   979 */ "GetTextureParameteriv",
    /*   980 */ "GetTransformFeedbacki64_v",
    /*   981 */ "GetTransformFeedbacki_v",
    /*   982 */ "GetTransformFeedbackiv",
    /*   983 */ "GetVertexArrayIndexed64iv",
    /*   984 */ "GetVertexArrayIndexediv",
    /*   985 */ "GetVertexArrayiv",
    /*   986 */ "InvalidateNamedFramebufferData",
    /*   987 */ "InvalidateNamedFramebufferSubData",
    /*   988 */ "MapNamedBuffer",
    /*   989 */ "MapNamedBufferRange",
    /*   990 */ "NamedBufferData",
    /*   991 */ "NamedBufferStorage",
    /*   992 */ "NamedBufferSubData",
    /*   993 */ "NamedFramebufferDrawBuffer",
    /*   994 */ "NamedFramebufferDrawBuffers",
    /*   995 */ "NamedFramebufferParameteri",
    /*   996 */ "NamedFramebufferReadBuffer",
    /*   997 */ "NamedFramebufferRenderbuffer",
    /*   998 */ "NamedFramebufferTexture",
    /*   999 */ "NamedFramebufferTextureLayer",
    /*  1000 */ "NamedRenderbufferStorage",
    /*  1001 */ "NamedRenderbufferStorageMultisample",
    /*  1002 */ "TextureBuffer",
    /*  1003 */ "TextureBufferRange",
    /*  1004 */ "TextureParameterIiv",
    /*  1005 */ "TextureParameterIuiv",
    /*  1006 */ "TextureParameterf",
    /*  1007 */ "TextureParameterfv",
    /*  1008 */ "TextureParameteri",
    /*  1009 */ "TextureParameteriv",
    /*  1010 */ "TextureStorage1D",
    /*  1011 */ "TextureStorage2D",
    /*  1012 */ "TextureStorage2DMultisample",
    /*  1013 */ "TextureStorage3D",
    /*  1014 */ "TextureStorage3DMultisample",
    /*  1015 */ "TextureSubImage1D",
    /*  1016 */ "TextureSubImage2D",
    /*  1017 */ "TextureSubImage3D",
    /*  1018 */ "TransformFeedbackBufferBase",
    /*  1019 */ "TransformFeedbackBufferRange",
    /*  1020 */ "UnmapNamedBufferEXT",
    /*  1021 */ "VertexArrayAttribBinding",
    /*  1022 */ "VertexArrayAttribFormat",
    /*  1023 */ "VertexArrayAttribIFormat",
    /*  1024 */ "VertexArrayAttribLFormat",
    /*  1025 */ "VertexArrayBindingDivisor",
    /*  1026 */ "VertexArrayElementBuffer",
    /*  1027 */ "VertexArrayVertexBuffer",
    /*  1028 */ "VertexArrayVertexBuffers",
    /*  1029 */ "GetCompressedTextureSubImage",
    /*  1030 */ "GetTextureSubImage",
    /*  1031 */ "BufferPageCommitmentARB",
    /*  1032 */ "NamedBufferPageCommitmentARB",
    /*  1033 */ "GetUniformi64vARB",
    /*  1034 */ "GetUniformui64vARB",
    /*  1035 */ "GetnUniformi64vARB",
    /*  1036 */ "GetnUniformui64vARB",
    /*  1037 */ "ProgramUniform1i64ARB",
    /*  1038 */ "ProgramUniform1i64vARB",
    /*  1039 */ "ProgramUniform1ui64ARB",
    /*  1040 */ "ProgramUniform1ui64vARB",
    /*  1041 */ "ProgramUniform2i64ARB",
    /*  1042 */ "ProgramUniform2i64vARB",
    /*  1043 */ "ProgramUniform2ui64ARB",
    /*  1044 */ "ProgramUniform2ui64vARB",
    /*  1045 */ "ProgramUniform3i64ARB",
    /*  1046 */ "ProgramUniform3i64vARB",
    /*  1047 */ "ProgramUniform3ui64ARB",
    /*  1048 */ "ProgramUniform3ui64vARB",
    /*  1049 */ "ProgramUniform4i64ARB",
    /*  1050 */ "ProgramUniform4i64vARB",
    /*  1051 */ "ProgramUniform4ui64ARB",
    /*  1052 */ "ProgramUniform4ui64vARB",
    /*  1053 */ "Uniform1i64ARB",
    /*  1054 */ "Uniform1i64vARB",
    /*  1055 */ "Uniform1ui64ARB",
    /*  1056 */ "Uniform1ui64vARB",
    /*  1057 */ "Uniform2i64ARB",
    /*  1058 */ "Uniform2i64vARB",
    /*  1059 */ "Uniform2ui64ARB",
    /*  1060 */ "Uniform2ui64vARB",
    /*  1061 */ "Uniform3i64ARB",
    /*  1062 */ "Uniform3i64vARB",
    /*  1063 */ "Uniform3ui64ARB",
    /*  1064 */ "Uniform3ui64vARB",
    /*  1065 */ "Uniform4i64ARB",
    /*  1066 */ "Uniform4i64vARB",
    /*  1067 */ "Uniform4ui64ARB",
    /*  1068 */ "Uniform4ui64vARB",
    /*  1069 */ "EvaluateDepthValuesARB",
    /*  1070 */ "FramebufferSampleLocationsfvARB",
    /*  1071 */ "NamedFramebufferSampleLocationsfvARB",
    /*  1072 */ "SpecializeShaderARB",
    /*  1073 */ "InvalidateBufferData",
    /*  1074 */ "InvalidateBufferSubData",
    /*  1075 */ "InvalidateFramebuffer",
    /*  1076 */ "InvalidateSubFramebuffer",
    /*  1077 */ "InvalidateTexImage",
    /*  1078 */ "InvalidateTexSubImage",
    /*  1079 */ "DrawTexfOES",
    /*  1080 */ "DrawTexfvOES",
    /*  1081 */ "DrawTexiOES",
    /*  1082 */ "DrawTexivOES",
    /*  1083 */ "DrawTexsOES",
    /*  1084 */ "DrawTexsvOES",
    /*  1085 */ "DrawTexxOES",
    /*  1086 */ "DrawTexxvOES",
    /*  1087 */ "PointSizePointerOES",
    /*  1088 */ "QueryMatrixxOES",
    /*  1089 */ "SampleMaskSGIS",
    /*  1090 */ "SamplePatternSGIS",
    /*  1091 */ "ColorPointerEXT",
    /*  1092 */ "EdgeFlagPointerEXT",
    /*  1093 */ "IndexPointerEXT",
    /*  1094 */ "NormalPointerEXT",
    /*  1095 */ "TexCoordPointerEXT",
    /*  1096 */ "VertexPointerEXT",
    /*  1097 */ "DiscardFramebufferEXT",
    /*  1098 */ "ActiveShaderProgram",
    /*  1099 */ "BindProgramPipeline",
    /*  1100 */ "CreateShaderProgramv",
    /*  1101 */ "DeleteProgramPipelines",
    /*  1102 */ "GenProgramPipelines",
    /*  1103 */ "GetProgramPipelineInfoLog",
    /*  1104 */ "GetProgramPipelineiv",
    /*  1105 */ "IsProgramPipeline",
    /*  1106 */ "LockArraysEXT",
    /*  1107 */ "ProgramUniform1d",
    /*  1108 */ "ProgramUniform1dv",
    /*  1109 */ "ProgramUniform1f",
    /*  1110 */ "ProgramUniform1fv",
    /*  1111 */ "ProgramUniform1i",
    /*  1112 */ "ProgramUniform1iv",
    /*  1113 */ "ProgramUniform1ui",
    /*  1114 */ "ProgramUniform1uiv",
    /*  1115 */ "ProgramUniform2d",
    /*  1116 */ "ProgramUniform2dv",
    /*  1117 */ "ProgramUniform2f",
    /*  1118 */ "ProgramUniform2fv",
    /*  1119 */ "ProgramUniform2i",
    /*  1120 */ "ProgramUniform2iv",
    /*  1121 */ "ProgramUniform2ui",
    /*  1122 */ "ProgramUniform2uiv",
    /*  1123 */ "ProgramUniform3d",
    /*  1124 */ "ProgramUniform3dv",
    /*  1125 */ "ProgramUniform3f",
    /*  1126 */ "ProgramUniform3fv",
    /*  1127 */ "ProgramUniform3i",
    /*  1128 */ "ProgramUniform3iv",
    /*  1129 */ "ProgramUniform3ui",
    /*  1130 */ "ProgramUniform3uiv",
    /*  1131 */ "ProgramUniform4d",
    /*  1132 */ "ProgramUniform4dv",
    /*  1133 */ "ProgramUniform4f",
    /*  1134 */ "ProgramUniform4fv",
    /*  1135 */ "ProgramUniform4i",
    /*  1136 */ "ProgramUniform4iv",
    /*  1137 */ "ProgramUniform4ui",
    /*  1138 */ "ProgramUniform4uiv",
    /*  1139 */ "ProgramUniformMatrix2dv",
    /*  1140 */ "ProgramUniformMatrix2fv",
    /*  1141 */ "ProgramUniformMatrix2x3dv",
    /*  1142 */ "ProgramUniformMatrix2x3fv",
    /*  1143 */ "ProgramUniformMatrix2x4dv",
    /*  1144 */ "ProgramUniformMatrix2x4fv",
    /*  1145 */ "ProgramUniformMatrix3dv",
    /*  1146 */ "ProgramUniformMatrix3fv",
    /*  1147 */ "ProgramUniformMatrix3x2dv",
    /*  1148 */ "ProgramUniformMatrix3x2fv",
    /*  1149 */ "ProgramUniformMatrix3x4dv",
    /*  1150 */ "ProgramUniformMatrix3x4fv",
    /*  1151 */ "ProgramUniformMatrix4dv",
    /*  1152 */ "ProgramUniformMatrix4fv",
    /*  1153 */ "ProgramUniformMatrix4x2dv",
    /*  1154 */ "ProgramUniformMatrix4x2fv",
    /*  1155 */ "ProgramUniformMatrix4x3dv",
    /*  1156 */ "ProgramUniformMatrix4x3fv",
    /*  1157 */ "UnlockArraysEXT",
    /*  1158 */ "UseProgramStages",
    /*  1159 */ "ValidateProgramPipeline",
    /*  1160 */ "FramebufferTexture2DMultisampleEXT",
    /*  1161 */ "DebugMessageCallback",
    /*  1162 */ "DebugMessageControl",
    /*  1163 */ "DebugMessageInsert",
    /*  1164 */ "GetDebugMessageLog",
    /*  1165 */ "GetObjectLabel",
    /*  1166 */ "GetObjectPtrLabel",
    /*  1167 */ "ObjectLabel",
    /*  1168 */ "ObjectPtrLabel",
    /*  1169 */ "PopDebugGroup",
    /*  1170 */ "PushDebugGroup",
    /*  1171 */ "SecondaryColor3fEXT",
    /*  1172 */ "SecondaryColor3fvEXT",
    /*  1173 */ "MultiDrawElements",
    /*  1174 */ "FogCoordfEXT",
    /*  1175 */ "FogCoordfvEXT",
    /*  1176 */ "ResizeBuffersMESA",
    /*  1177 */ "WindowPos4dMESA",
    /*  1178 */ "WindowPos4dvMESA",
    /*  1179 */ "WindowPos4fMESA",
    /*  1180 */ "WindowPos4fvMESA",
    /*  1181 */ "WindowPos4iMESA",
    /*  1182 */ "WindowPos4ivMESA",
    /*  1183 */ "WindowPos4sMESA",
    /*  1184 */ "WindowPos4svMESA",
    /*  1185 */ "MultiModeDrawArraysIBM",
    /*  1186 */ "MultiModeDrawElementsIBM",
    /*  1187 */ "AreProgramsResidentNV",
    /*  1188 */ "ExecuteProgramNV",
    /*  1189 */ "GetProgramParameterdvNV",
    /*  1190 */ "GetProgramParameterfvNV",
    /*  1191 */ "GetProgramStringNV",
    /*  1192 */ "GetProgramivNV",
    /*  1193 */ "GetTrackMatrixivNV",
    /*  1194 */ "GetVertexAttribdvNV",
    /*  1195 */ "GetVertexAttribfvNV",
    /*  1196 */ "GetVertexAttribivNV",
    /*  1197 */ "LoadProgramNV",
    /*  1198 */ "ProgramParameters4dvNV",
    /*  1199 */ "ProgramParameters4fvNV",
    /*  1200 */ "RequestResidentProgramsNV",
    /*  1201 */ "TrackMatrixNV",
    /*  1202 */ "VertexAttrib1dNV",
    /*  1203 */ "VertexAttrib1dvNV",
    /*  1204 */ "VertexAttrib1fNV",
    /*  1205 */ "VertexAttrib1fvNV",
    /*  1206 */ "VertexAttrib1sNV",
    /*  1207 */ "VertexAttrib1svNV",
    /*  1208 */ "VertexAttrib2dNV",
    /*  1209 */ "VertexAttrib2dvNV",
    /*  1210 */ "VertexAttrib2fNV",
    /*  1211 */ "VertexAttrib2fvNV",
    /*  1212 */ "VertexAttrib2sNV",
    /*  1213 */ "VertexAttrib2svNV",
    /*  1214 */ "VertexAttrib3dNV",
    /*  1215 */ "VertexAttrib3dvNV",
    /*  1216 */ "VertexAttrib3fNV",
    /*  1217 */ "VertexAttrib3fvNV",
    /*  1218 */ "VertexAttrib3sNV",
    /*  1219 */ "VertexAttrib3svNV",
    /*  1220 */ "VertexAttrib4dNV",
    /*  1221 */ "VertexAttrib4dvNV",
    /*  1222 */ "VertexAttrib4fNV",
    /*  1223 */ "VertexAttrib4fvNV",
    /*  1224 */ "VertexAttrib4sNV",
    /*  1225 */ "VertexAttrib4svNV",
    /*  1226 */ "VertexAttrib4ubNV",
    /*  1227 */ "VertexAttrib4ubvNV",
    /*  1228 */ "VertexAttribPointerNV",
    /*  1229 */ "VertexAttribs1dvNV",
    /*  1230 */ "VertexAttribs1fvNV",
    /*  1231 */ "VertexAttribs1svNV",
    /*  1232 */ "VertexAttribs2dvNV",
    /*  1233 */ "VertexAttribs2fvNV",
    /*  1234 */ "VertexAttribs2svNV",
    /*  1235 */ "VertexAttribs3dvNV",
    /*  1236 */ "VertexAttribs3fvNV",
    /*  1237 */ "VertexAttribs3svNV",
    /*  1238 */ "VertexAttribs4dvNV",
    /*  1239 */ "VertexAttribs4fvNV",
    /*  1240 */ "VertexAttribs4svNV",
    /*  1241 */ "VertexAttribs4ubvNV",
    /*  1242 */ "GetTexBumpParameterfvATI",
    /*  1243 */ "GetTexBumpParameterivATI",
    /*  1244 */ "TexBumpParameterfvATI",
    /*  1245 */ "TexBumpParameterivATI",
    /*  1246 */ "AlphaFragmentOp1ATI",
    /*  1247 */ "AlphaFragmentOp2ATI",
    /*  1248 */ "AlphaFragmentOp3ATI",
    /*  1249 */ "BeginFragmentShaderATI",
    /*  1250 */ "BindFragmentShaderATI",
    /*  1251 */ "ColorFragmentOp1ATI",
    /*  1252 */ "ColorFragmentOp2ATI",
    /*  1253 */ "ColorFragmentOp3ATI",
    /*  1254 */ "DeleteFragmentShaderATI",
    /*  1255 */ "EndFragmentShaderATI",
    /*  1256 */ "GenFragmentShadersATI",
    /*  1257 */ "PassTexCoordATI",
    /*  1258 */ "SampleMapATI",
    /*  1259 */ "SetFragmentShaderConstantATI",
    /*  1260 */ "DepthRangeArrayfvOES",
    /*  1261 */ "DepthRangeIndexedfOES",
    /*  1262 */ "ActiveStencilFaceEXT",
    /*  1263 */ "GetProgramNamedParameterdvNV",
    /*  1264 */ "GetProgramNamedParameterfvNV",
    /*  1265 */ "ProgramNamedParameter4dNV",
    /*  1266 */ "ProgramNamedParameter4dvNV",
    /*  1267 */ "ProgramNamedParameter4fNV",
    /*  1268 */ "ProgramNamedParameter4fvNV",
    /*  1269 */ "PrimitiveRestartNV",
    /*  1270 */ "GetTexGenxvOES",
    /*  1271 */ "TexGenxOES",
    /*  1272 */ "TexGenxvOES",
    /*  1273 */ "DepthBoundsEXT",
    /*  1274 */ "BindFramebufferEXT",
    /*  1275 */ "BindRenderbufferEXT",
    /*  1276 */ "StringMarkerGREMEDY",
    /*  1277 */ "BufferParameteriAPPLE",
    /*  1278 */ "FlushMappedBufferRangeAPPLE",
    /*  1279 */ "VertexAttribI1iEXT",
    /*  1280 */ "VertexAttribI1uiEXT",
    /*  1281 */ "VertexAttribI2iEXT",
    /*  1282 */ "VertexAttribI2ivEXT",
    /*  1283 */ "VertexAttribI2uiEXT",
    /*  1284 */ "VertexAttribI2uivEXT",
    /*  1285 */ "VertexAttribI3iEXT",
    /*  1286 */ "VertexAttribI3ivEXT",
    /*  1287 */ "VertexAttribI3uiEXT",
    /*  1288 */ "VertexAttribI3uivEXT",
    /*  1289 */ "VertexAttribI4iEXT",
    /*  1290 */ "VertexAttribI4ivEXT",
    /*  1291 */ "VertexAttribI4uiEXT",
    /*  1292 */ "VertexAttribI4uivEXT",
    /*  1293 */ "ClearColorIiEXT",
    /*  1294 */ "ClearColorIuiEXT",
    /*  1295 */ "BindBufferOffsetEXT",
    /*  1296 */ "BeginPerfMonitorAMD",
    /*  1297 */ "DeletePerfMonitorsAMD",
    /*  1298 */ "EndPerfMonitorAMD",
    /*  1299 */ "GenPerfMonitorsAMD",
    /*  1300 */ "GetPerfMonitorCounterDataAMD",
    /*  1301 */ "GetPerfMonitorCounterInfoAMD",
    /*  1302 */ "GetPerfMonitorCounterStringAMD",
    /*  1303 */ "GetPerfMonitorCountersAMD",
    /*  1304 */ "GetPerfMonitorGroupStringAMD",
    /*  1305 */ "GetPerfMonitorGroupsAMD",
    /*  1306 */ "SelectPerfMonitorCountersAMD",
    /*  1307 */ "GetObjectParameterivAPPLE",
    /*  1308 */ "ObjectPurgeableAPPLE",
    /*  1309 */ "ObjectUnpurgeableAPPLE",
    /*  1310 */ "ActiveProgramEXT",
    /*  1311 */ "CreateShaderProgramEXT",
    /*  1312 */ "UseShaderProgramEXT",
    /*  1313 */ "TextureBarrierNV",
    /*  1314 */ "VDPAUFiniNV",
    /*  1315 */ "VDPAUGetSurfaceivNV",
    /*  1316 */ "VDPAUInitNV",
    /*  1317 */ "VDPAUIsSurfaceNV",
    /*  1318 */ "VDPAUMapSurfacesNV",
    /*  1319 */ "VDPAURegisterOutputSurfaceNV",
    /*  1320 */ "VDPAURegisterVideoSurfaceNV",
    /*  1321 */ "VDPAUSurfaceAccessNV",
    /*  1322 */ "VDPAUUnmapSurfacesNV",
    /*  1323 */ "VDPAUUnregisterSurfaceNV",
    /*  1324 */ "BeginPerfQueryINTEL",
    /*  1325 */ "CreatePerfQueryINTEL",
    /*  1326 */ "DeletePerfQueryINTEL",
    /*  1327 */ "EndPerfQueryINTEL",
    /*  1328 */ "GetFirstPerfQueryIdINTEL",
    /*  1329 */ "GetNextPerfQueryIdINTEL",
    /*  1330 */ "GetPerfCounterInfoINTEL",
    /*  1331 */ "GetPerfQueryDataINTEL",
    /*  1332 */ "GetPerfQueryIdByNameINTEL",
    /*  1333 */ "GetPerfQueryInfoINTEL",
    /*  1334 */ "PolygonOffsetClampEXT",
    /*  1335 */ "SubpixelPrecisionBiasNV",
    /*  1336 */ "ConservativeRasterParameterfNV",
    /*  1337 */ "ConservativeRasterParameteriNV",
    /*  1338 */ "WindowRectanglesEXT",
    /*  1339 */ "BufferStorageMemEXT",
    /*  1340 */ "CreateMemoryObjectsEXT",
    /*  1341 */ "DeleteMemoryObjectsEXT",
    /*  1342 */ "DeleteSemaphoresEXT",
    /*  1343 */ "GenSemaphoresEXT",
    /*  1344 */ "GetMemoryObjectParameterivEXT",
    /*  1345 */ "GetSemaphoreParameterui64vEXT",
    /*  1346 */ "GetUnsignedBytei_vEXT",
    /*  1347 */ "GetUnsignedBytevEXT",
    /*  1348 */ "IsMemoryObjectEXT",
    /*  1349 */ "IsSemaphoreEXT",
    /*  1350 */ "MemoryObjectParameterivEXT",
    /*  1351 */ "NamedBufferStorageMemEXT",
    /*  1352 */ "SemaphoreParameterui64vEXT",
    /*  1353 */ "SignalSemaphoreEXT",
    /*  1354 */ "TexStorageMem1DEXT",
    /*  1355 */ "TexStorageMem2DEXT",
    /*  1356 */ "TexStorageMem2DMultisampleEXT",
    /*  1357 */ "TexStorageMem3DEXT",
    /*  1358 */ "TexStorageMem3DMultisampleEXT",
    /*  1359 */ "TextureStorageMem1DEXT",
    /*  1360 */ "TextureStorageMem2DEXT",
    /*  1361 */ "TextureStorageMem2DMultisampleEXT",
    /*  1362 */ "TextureStorageMem3DEXT",
    /*  1363 */ "TextureStorageMem3DMultisampleEXT",
    /*  1364 */ "WaitSemaphoreEXT",
    /*  1365 */ "ImportMemoryFdEXT",
    /*  1366 */ "ImportSemaphoreFdEXT",
    /*  1367 */ "FramebufferFetchBarrierEXT",
    /*  1368 */ "NamedRenderbufferStorageMultisampleAdvancedAMD",
    /*  1369 */ "RenderbufferStorageMultisampleAdvancedAMD",
    /*  1370 */ "StencilFuncSeparateATI",
    /*  1371 */ "ProgramEnvParameters4fvEXT",
    /*  1372 */ "ProgramLocalParameters4fvEXT",
    /*  1373 */ "EGLImageTargetRenderbufferStorageOES",
    /*  1374 */ "EGLImageTargetTexture2DOES",
    /*  1375 */ "AlphaFuncx",
    /*  1376 */ "ClearColorx",
    /*  1377 */ "ClearDepthx",
    /*  1378 */ "Color4x",
    /*  1379 */ "DepthRangex",
    /*  1380 */ "Fogx",
    /*  1381 */ "Fogxv",
    /*  1382 */ "Frustumf",
    /*  1383 */ "Frustumx",
    /*  1384 */ "LightModelx",
    /*  1385 */ "LightModelxv",
    /*  1386 */ "Lightx",
    /*  1387 */ "Lightxv",
    /*  1388 */ "LineWidthx",
    /*  1389 */ "LoadMatrixx",
    /*  1390 */ "Materialx",
    /*  1391 */ "Materialxv",
    /*  1392 */ "MultMatrixx",
    /*  1393 */ "MultiTexCoord4x",
    /*  1394 */ "Normal3x",
    /*  1395 */ "Orthof",
    /*  1396 */ "Orthox",
    /*  1397 */ "PointSizex",
    /*  1398 */ "PolygonOffsetx",
    /*  1399 */ "Rotatex",
    /*  1400 */ "SampleCoveragex",
    /*  1401 */ "Scalex",
    /*  1402 */ "TexEnvx",
    /*  1403 */ "TexEnvxv",
    /*  1404 */ "TexParameterx",
    /*  1405 */ "Translatex",
    /*  1406 */ "ClipPlanef",
    /*  1407 */ "ClipPlanex",
    /*  1408 */ "GetClipPlanef",
    /*  1409 */ "GetClipPlanex",
    /*  1410 */ "GetFixedv",
    /*  1411 */ "GetLightxv",
    /*  1412 */ "GetMaterialxv",
    /*  1413 */ "GetTexEnvxv",
    /*  1414 */ "GetTexParameterxv",
    /*  1415 */ "PointParameterx",
    /*  1416 */ "PointParameterxv",
    /*  1417 */ "TexParameterxv",
    /*  1418 */ "BlendBarrier",
    /*  1419 */ "PrimitiveBoundingBox",
    /*  1420 */ "MaxShaderCompilerThreadsKHR",
    /*  1421 */ "MatrixLoadfEXT",
    /*  1422 */ "MatrixLoaddEXT",
    /*  1423 */ "MatrixMultfEXT",
    /*  1424 */ "MatrixMultdEXT",
    /*  1425 */ "MatrixLoadIdentityEXT",
    /*  1426 */ "MatrixRotatefEXT",
    /*  1427 */ "MatrixRotatedEXT",
    /*  1428 */ "MatrixScalefEXT",
    /*  1429 */ "MatrixScaledEXT",
    /*  1430 */ "MatrixTranslatefEXT",
    /*  1431 */ "MatrixTranslatedEXT",
    /*  1432 */ "MatrixOrthoEXT",
    /*  1433 */ "MatrixFrustumEXT",
    /*  1434 */ "MatrixPushEXT",
    /*  1435 */ "MatrixPopEXT",
    /*  1436 */ "MatrixLoadTransposefEXT",
    /*  1437 */ "MatrixLoadTransposedEXT",
    /*  1438 */ "MatrixMultTransposefEXT",
    /*  1439 */ "MatrixMultTransposedEXT",
    /*  1440 */ "BindMultiTextureEXT",
    /*  1441 */ "NamedBufferDataEXT",
    /*  1442 */ "NamedBufferSubDataEXT",
    /*  1443 */ "NamedBufferStorageEXT",
    /*  1444 */ "MapNamedBufferRangeEXT",
    /*  1445 */ "TextureImage1DEXT",
    /*  1446 */ "TextureImage2DEXT",
    /*  1447 */ "TextureImage3DEXT",
    /*  1448 */ "TextureSubImage1DEXT",
    /*  1449 */ "TextureSubImage2DEXT",
    /*  1450 */ "TextureSubImage3DEXT",
    /*  1451 */ "CopyTextureImage1DEXT",
    /*  1452 */ "CopyTextureImage2DEXT",
    /*  1453 */ "CopyTextureSubImage1DEXT",
    /*  1454 */ "CopyTextureSubImage2DEXT",
    /*  1455 */ "CopyTextureSubImage3DEXT",
    /*  1456 */ "MapNamedBufferEXT",
    /*  1457 */ "GetTextureParameterivEXT",
    /*  1458 */ "GetTextureParameterfvEXT",
    /*  1459 */ "TextureParameteriEXT",
    /*  1460 */ "TextureParameterivEXT",
    /*  1461 */ "TextureParameterfEXT",
    /*  1462 */ "TextureParameterfvEXT",
    /*  1463 */ "GetTextureImageEXT",
    /*  1464 */ "GetTextureLevelParameterivEXT",
    /*  1465 */ "GetTextureLevelParameterfvEXT",
    /*  1466 */ "GetNamedBufferSubDataEXT",
    /*  1467 */ "GetNamedBufferPointervEXT",
    /*  1468 */ "GetNamedBufferParameterivEXT",
    /*  1469 */ "FlushMappedNamedBufferRangeEXT",
    /*  1470 */ "FramebufferDrawBufferEXT",
    /*  1471 */ "FramebufferDrawBuffersEXT",
    /*  1472 */ "FramebufferReadBufferEXT",
    /*  1473 */ "GetFramebufferParameterivEXT",
    /*  1474 */ "CheckNamedFramebufferStatusEXT",
    /*  1475 */ "NamedFramebufferTexture1DEXT",
    /*  1476 */ "NamedFramebufferTexture2DEXT",
    /*  1477 */ "NamedFramebufferTexture3DEXT",
    /*  1478 */ "NamedFramebufferRenderbufferEXT",
    /*  1479 */ "GetNamedFramebufferAttachmentParameterivEXT",
    /*  1480 */ "EnableClientStateiEXT",
    /*  1481 */ "DisableClientStateiEXT",
    /*  1482 */ "GetPointerIndexedvEXT",
    /*  1483 */ "MultiTexEnviEXT",
    /*  1484 */ "MultiTexEnvivEXT",
    /*  1485 */ "MultiTexEnvfEXT",
    /*  1486 */ "MultiTexEnvfvEXT",
    /*  1487 */ "GetMultiTexEnvivEXT",
    /*  1488 */ "GetMultiTexEnvfvEXT",
    /*  1489 */ "MultiTexParameteriEXT",
    /*  1490 */ "MultiTexParameterivEXT",
    /*  1491 */ "MultiTexParameterfEXT",
    /*  1492 */ "MultiTexParameterfvEXT",
    /*  1493 */ "GetMultiTexImageEXT",
    /*  1494 */ "MultiTexImage1DEXT",
    /*  1495 */ "MultiTexImage2DEXT",
    /*  1496 */ "MultiTexImage3DEXT",
    /*  1497 */ "MultiTexSubImage1DEXT",
    /*  1498 */ "MultiTexSubImage2DEXT",
    /*  1499 */ "MultiTexSubImage3DEXT",
    /*  1500 */ "GetMultiTexParameterivEXT",
    /*  1501 */ "GetMultiTexParameterfvEXT",
    /*  1502 */ "CopyMultiTexImage1DEXT",
    /*  1503 */ "CopyMultiTexImage2DEXT",
    /*  1504 */ "CopyMultiTexSubImage1DEXT",
    /*  1505 */ "CopyMultiTexSubImage2DEXT",
    /*  1506 */ "CopyMultiTexSubImage3DEXT",
    /*  1507 */ "MultiTexGendEXT",
    /*  1508 */ "MultiTexGendvEXT",
    /*  1509 */ "MultiTexGenfEXT",
    /*  1510 */ "MultiTexGenfvEXT",
    /*  1511 */ "MultiTexGeniEXT",
    /*  1512 */ "MultiTexGenivEXT",
    /*  1513 */ "GetMultiTexGendvEXT",
    /*  1514 */ "GetMultiTexGenfvEXT",
    /*  1515 */ "GetMultiTexGenivEXT",
    /*  1516 */ "MultiTexCoordPointerEXT",
    /*  1517 */ "BindImageTextureEXT",
    /*  1518 */ "CompressedTextureImage1DEXT",
    /*  1519 */ "CompressedTextureImage2DEXT",
    /*  1520 */ "CompressedTextureImage3DEXT",
    /*  1521 */ "CompressedTextureSubImage1DEXT",
    /*  1522 */ "CompressedTextureSubImage2DEXT",
    /*  1523 */ "CompressedTextureSubImage3DEXT",
    /*  1524 */ "GetCompressedTextureImageEXT",
    /*  1525 */ "CompressedMultiTexImage1DEXT",
    /*  1526 */ "CompressedMultiTexImage2DEXT",
    /*  1527 */ "CompressedMultiTexImage3DEXT",
    /*  1528 */ "CompressedMultiTexSubImage1DEXT",
    /*  1529 */ "CompressedMultiTexSubImage2DEXT",
    /*  1530 */ "CompressedMultiTexSubImage3DEXT",
    /*  1531 */ "GetCompressedMultiTexImageEXT",
    /*  1532 */ "GetMultiTexLevelParameterivEXT",
    /*  1533 */ "GetMultiTexLevelParameterfvEXT",
    /*  1534 */ "FramebufferParameteriMESA",
    /*  1535 */ "GetFramebufferParameterivMESA",
    /*  1536 */ "NamedRenderbufferStorageEXT",
    /*  1537 */ "GetNamedRenderbufferParameterivEXT",
    /*  1538 */ "ClientAttribDefaultEXT",
    /*  1539 */ "PushClientAttribDefaultEXT",
    /*  1540 */ "NamedProgramStringEXT",
    /*  1541 */ "GetNamedProgramStringEXT",
    /*  1542 */ "NamedProgramLocalParameter4fEXT",
    /*  1543 */ "NamedProgramLocalParameter4fvEXT",
    /*  1544 */ "GetNamedProgramLocalParameterfvEXT",
    /*  1545 */ "NamedProgramLocalParameter4dEXT",
    /*  1546 */ "NamedProgramLocalParameter4dvEXT",
    /*  1547 */ "GetNamedProgramLocalParameterdvEXT",
    /*  1548 */ "GetNamedProgramivEXT",
    /*  1549 */ "TextureBufferEXT",
    /*  1550 */ "MultiTexBufferEXT",
    /*  1551 */ "TextureParameterIivEXT",
    /*  1552 */ "TextureParameterIuivEXT",
    /*  1553 */ "GetTextureParameterIivEXT",
    /*  1554 */ "GetTextureParameterIuivEXT",
    /*  1555 */ "MultiTexParameterIivEXT",
    /*  1556 */ "MultiTexParameterIuivEXT",
    /*  1557 */ "GetMultiTexParameterIivEXT",
    /*  1558 */ "GetMultiTexParameterIuivEXT",
    /*  1559 */ "NamedProgramLocalParameters4fvEXT",
    /*  1560 */ "GenerateTextureMipmapEXT",
    /*  1561 */ "GenerateMultiTexMipmapEXT",
    /*  1562 */ "NamedRenderbufferStorageMultisampleEXT",
    /*  1563 */ "NamedCopyBufferSubDataEXT",
    /*  1564 */ "VertexArrayVertexOffsetEXT",
    /*  1565 */ "VertexArrayColorOffsetEXT",
    /*  1566 */ "VertexArrayEdgeFlagOffsetEXT",
    /*  1567 */ "VertexArrayIndexOffsetEXT",
    /*  1568 */ "VertexArrayNormalOffsetEXT",
    /*  1569 */ "VertexArrayTexCoordOffsetEXT",
    /*  1570 */ "VertexArrayMultiTexCoordOffsetEXT",
    /*  1571 */ "VertexArrayFogCoordOffsetEXT",
    /*  1572 */ "VertexArraySecondaryColorOffsetEXT",
    /*  1573 */ "VertexArrayVertexAttribOffsetEXT",
    /*  1574 */ "VertexArrayVertexAttribIOffsetEXT",
    /*  1575 */ "EnableVertexArrayEXT",
    /*  1576 */ "DisableVertexArrayEXT",
    /*  1577 */ "EnableVertexArrayAttribEXT",
    /*  1578 */ "DisableVertexArrayAttribEXT",
    /*  1579 */ "GetVertexArrayIntegervEXT",
    /*  1580 */ "GetVertexArrayPointervEXT",
    /*  1581 */ "GetVertexArrayIntegeri_vEXT",
    /*  1582 */ "GetVertexArrayPointeri_vEXT",
    /*  1583 */ "ClearNamedBufferDataEXT",
    /*  1584 */ "ClearNamedBufferSubDataEXT",
    /*  1585 */ "NamedFramebufferParameteriEXT",
    /*  1586 */ "GetNamedFramebufferParameterivEXT",
    /*  1587 */ "VertexArrayVertexAttribLOffsetEXT",
    /*  1588 */ "VertexArrayVertexAttribDivisorEXT",
    /*  1589 */ "TextureBufferRangeEXT",
    /*  1590 */ "TextureStorage2DMultisampleEXT",
    /*  1591 */ "TextureStorage3DMultisampleEXT",
    /*  1592 */ "VertexArrayBindVertexBufferEXT",
    /*  1593 */ "VertexArrayVertexAttribFormatEXT",
    /*  1594 */ "VertexArrayVertexAttribIFormatEXT",
    /*  1595 */ "VertexArrayVertexAttribLFormatEXT",
    /*  1596 */ "VertexArrayVertexAttribBindingEXT",
    /*  1597 */ "VertexArrayVertexBindingDivisorEXT",
    /*  1598 */ "NamedBufferPageCommitmentEXT",
    /*  1599 */ "NamedStringARB",
    /*  1600 */ "DeleteNamedStringARB",
    /*  1601 */ "CompileShaderIncludeARB",
    /*  1602 */ "IsNamedStringARB",
    /*  1603 */ "GetNamedStringARB",
    /*  1604 */ "GetNamedStringivARB",
    /*  1605 */ "EGLImageTargetTexStorageEXT",
    /*  1606 */ "EGLImageTargetTextureStorageEXT",
    /*  1607 */ "CopyImageSubDataNV",
    /*  1608 */ "ViewportSwizzleNV",
    /*  1609 */ "AlphaToCoverageDitherControlNV",
    /*  1610 */ "InternalBufferSubDataCopyMESA",
    /*  1611 */ "Vertex2hNV",
    /*  1612 */ "Vertex2hvNV",
    /*  1613 */ "Vertex3hNV",
    /*  1614 */ "Vertex3hvNV",
    /*  1615 */ "Vertex4hNV",
    /*  1616 */ "Vertex4hvNV",
    /*  1617 */ "Normal3hNV",
    /*  1618 */ "Normal3hvNV",
    /*  1619 */ "Color3hNV",
    /*  1620 */ "Color3hvNV",
    /*  1621 */ "Color4hNV",
    /*  1622 */ "Color4hvNV",
    /*  1623 */ "TexCoord1hNV",
    /*  1624 */ "TexCoord1hvNV",
    /*  1625 */ "TexCoord2hNV",
    /*  1626 */ "TexCoord2hvNV",
    /*  1627 */ "TexCoord3hNV",
    /*  1628 */ "TexCoord3hvNV",
    /*  1629 */ "TexCoord4hNV",
    /*  1630 */ "TexCoord4hvNV",
    /*  1631 */ "MultiTexCoord1hNV",
    /*  1632 */ "MultiTexCoord1hvNV",
    /*  1633 */ "MultiTexCoord2hNV",
    /*  1634 */ "MultiTexCoord2hvNV",
    /*  1635 */ "MultiTexCoord3hNV",
    /*  1636 */ "MultiTexCoord3hvNV",
    /*  1637 */ "MultiTexCoord4hNV",
    /*  1638 */ "MultiTexCoord4hvNV",
    /*  1639 */ "FogCoordhNV",
    /*  1640 */ "FogCoordhvNV",
    /*  1641 */ "SecondaryColor3hNV",
    /*  1642 */ "SecondaryColor3hvNV",
    /*  1643 */ "InternalSetError",
    /*  1644 */ "VertexAttrib1hNV",
    /*  1645 */ "VertexAttrib1hvNV",
    /*  1646 */ "VertexAttrib2hNV",
    /*  1647 */ "VertexAttrib2hvNV",
    /*  1648 */ "VertexAttrib3hNV",
    /*  1649 */ "VertexAttrib3hvNV",
    /*  1650 */ "VertexAttrib4hNV",
    /*  1651 */ "VertexAttrib4hvNV",
    /*  1652 */ "VertexAttribs1hvNV",
    /*  1653 */ "VertexAttribs2hvNV",
    /*  1654 */ "VertexAttribs3hvNV",
    /*  1655 */ "VertexAttribs4hvNV",
    /*  1656 */ "TexPageCommitmentARB",
    /*  1657 */ "TexturePageCommitmentEXT",
    /*  1658 */ "ImportMemoryWin32HandleEXT",
    /*  1659 */ "ImportSemaphoreWin32HandleEXT",
    /*  1660 */ "ImportMemoryWin32NameEXT",
    /*  1661 */ "ImportSemaphoreWin32NameEXT",
    /*  1662 */ "GetObjectLabelEXT",
    /*  1663 */ "LabelObjectEXT",
    /*  1664 */ "DrawArraysUserBuf",
    /*  1665 */ "DrawElementsUserBuf",
    /*  1666 */ "MultiDrawArraysUserBuf",
    /*  1667 */ "MultiDrawElementsUserBuf",
    /*  1668 */ "DrawArraysInstancedBaseInstanceDrawID",
    /*  1669 */ "DrawElementsInstancedBaseVertexBaseInstanceDrawID",
    /*  1670 */ "InternalInvalidateFramebufferAncillaryMESA",
};

struct _glapi_table *
_glapi_create_table_from_handle(void *handle, const char *symbol_prefix) {
    struct _glapi_table *disp = calloc(_glapi_get_dispatch_table_size(), sizeof(_glapi_proc));
    char symboln[512];

    if(!disp)
        return NULL;

    if(symbol_prefix == NULL)
        symbol_prefix = "";

    /* Note: This code relies on _glapi_table_func_names being sorted by the
     * entry point index of each function.
     */
    for (int func_index = 0; func_index < GLAPI_TABLE_COUNT; ++func_index) {
        const char *name = _glapi_table_func_names[func_index];
        void ** procp = &((void **)disp)[func_index];

        snprintf(symboln, sizeof(symboln), "%s%s", symbol_prefix, name);
#ifdef _WIN32
        *procp = GetProcAddress(handle, symboln);
#else
        *procp = dlsym(handle, symboln);
#endif
    }
    __glapi_gentable_set_remaining_noop(disp);

    return disp;
}

void
 _glapi_table_patch(struct _glapi_table *table, const char *name, void *wrapper)
{
   for (int func_index = 0; func_index < GLAPI_TABLE_COUNT; ++func_index) {
      if (!strcmp(_glapi_table_func_names[func_index], name)) {
            ((void **)table)[func_index] = wrapper;
            return;
         }
   }
   fprintf(stderr, "could not patch %s in dispatch table\n", name);
}


